#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <richedit.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "Resources\Resource.h"


//Defines
#define STRID_OK          1
#define STRID_CANCEL      2
#define STRID_SYMBOL      3
#define STRID_COLOR       4
#define STRID_SELCOLOR    5
#define STRID_SPACE       6
#define STRID_TAB         7
#define STRID_NEWLINE     8
#define STRID_VERTICALTAB 9
#define STRID_FORMFEED    10
#define STRID_NULL        11

typedef struct _SPECIALCOLOR {
  DWORD dwColor;
  DWORD dwSelColor;
  BOOL bColorEnable;
  BOOL bSelColorEnable;
} SPECIALCOLOR;

typedef struct _SPECIALCHARA {
  char szNewChar[4];
  char szOldChar[4];
  SPECIALCOLOR sclr;
} SPECIALCHARA;

typedef struct _SPECIALCHARW {
  wchar_t wszNewChar[4];
  wchar_t wszOldChar[4];
  SPECIALCOLOR sclr;
} SPECIALCHARW;


//Functions prototypes
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void UpdateEdit(HWND hWnd);
void UpdateEditAll();
void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt, TEXTMETRICA *tm);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA();
void SaveOptionsW();
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hMdiClient;
BOOL bOldWindows;
BOOL bOldRichEdit;
BOOL bAkelEdit;
BOOL bMDI;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
SPECIALCHARA scSpaceA;
SPECIALCHARA scTabA;
SPECIALCHARA scNewLineA;
SPECIALCHARA scVerticalTabA;
SPECIALCHARA scFormFeedA;
SPECIALCHARA scNullA;
SPECIALCHARW scSpaceW;
SPECIALCHARW scTabW;
SPECIALCHARW scNewLineW;
SPECIALCHARW scVerticalTabW;
SPECIALCHARW scFormFeedW;
SPECIALCHARW scNullW;
WNDPROCDATA *NewEditProcData=NULL;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="SpecialChar";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bOldWindows)
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcA);
  else
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if (pd->bInMemory) pd->nUnload=UD_NONUNLOAD_NONACTIVE;
}

extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  //Is plugin already loaded?
  if (bInitMain)
  {
    UninitMain();

    UpdateEditAll();
  }
  else
  {
    InitMain();

    if (!pd->bOnStart) UpdateEditAll();

    //Stay in memory, and show as active
    pd->nUnload=UD_NONUNLOAD_ACTIVE;
  }
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HICON hPluginIcon;
  static SPECIALCHARA scSpaceDlgA;
  static SPECIALCHARA scTabDlgA;
  static SPECIALCHARA scNewLineDlgA;
  static SPECIALCHARA scVerticalTabDlgA;
  static SPECIALCHARA scFormFeedDlgA;
  static SPECIALCHARA scNullDlgA;
  static SPECIALCHARA *pscCharA;
  static HWND hWndList;
  static HWND hWndCharEdit;
  static HWND hWndColorCheck;
  static HWND hWndColorButton;
  static HWND hWndColorLabel;
  static HWND hWndSelColorCheck;
  static HWND hWndSelColorButton;
  static HWND hWndSelColorLabel;

  if (uMsg == WM_INITDIALOG)
  {
    hPluginIcon=LoadIconA(hInstanceDLL, MAKEINTRESOURCEA(IDI_ICON));
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hPluginIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);
    hWndCharEdit=GetDlgItem(hDlg, IDC_CHAR_EDIT);
    hWndColorCheck=GetDlgItem(hDlg, IDC_COLOR_CHECK);
    hWndColorButton=GetDlgItem(hDlg, IDC_COLOR_BUTTON);
    hWndColorLabel=GetDlgItem(hDlg, IDC_COLOR_LABEL);
    hWndSelColorCheck=GetDlgItem(hDlg, IDC_SELCOLOR_CHECK);
    hWndSelColorButton=GetDlgItem(hDlg, IDC_SELCOLOR_BUTTON);
    hWndSelColorLabel=GetDlgItem(hDlg, IDC_SELCOLOR_LABEL);

    SetDlgItemTextA(hDlg, IDC_CHAR_LABEL, GetLangStringA(wLangSystem, STRID_SYMBOL));
    SetDlgItemTextA(hDlg, IDC_COLOR_LABEL, GetLangStringA(wLangSystem, STRID_COLOR));
    SetDlgItemTextA(hDlg, IDC_SELCOLOR_LABEL, GetLangStringA(wLangSystem, STRID_SELCOLOR));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    scSpaceDlgA=scSpaceA;
    scTabDlgA=scTabA;
    scNewLineDlgA=scNewLineA;
    scVerticalTabDlgA=scVerticalTabA;
    scFormFeedDlgA=scFormFeedA;
    scNullDlgA=scNullA;
    pscCharA=&scSpaceDlgA;

    SendMessage(hWndCharEdit, EM_LIMITTEXT, 1, 0);
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_SPACE));
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_TAB));
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_NEWLINE));
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_VERTICALTAB));
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_FORMFEED));
    SendMessageA(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringA(wLangSystem, STRID_NULL));
    SendMessage(hWndList, LB_SETCURSEL, 0, 0);
    SendMessage(hDlg, WM_COMMAND, MAKELONG(IDC_LIST, LBN_SELCHANGE), (LPARAM)hWndList);
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_COLOR_BUTTON)
    {
      if (pscCharA->sclr.bColorEnable)
        dwColor=pscCharA->sclr.dwColor;
    }
    else if (wParam == IDC_SELCOLOR_BUTTON)
    {
      if (pscCharA->sclr.bSelColorEnable)
        dwColor=pscCharA->sclr.dwSelColor;
    }

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_LIST)
    {
      if (HIWORD(wParam) == LBN_SELCHANGE)
      {
        int nItem;

        nItem=SendMessage(hWndList, LB_GETCURSEL, 0, 0);

        if (nItem == 0)
          pscCharA=&scSpaceDlgA;
        else if (nItem == 1)
          pscCharA=&scTabDlgA;
        else if (nItem == 2)
          pscCharA=&scNewLineDlgA;
        else if (nItem == 3)
          pscCharA=&scVerticalTabDlgA;
        else if (nItem == 4)
          pscCharA=&scFormFeedDlgA;
        else if (nItem == 5)
          pscCharA=&scNullDlgA;

        SetWindowTextA(hWndCharEdit, pscCharA->szNewChar);

        if (pscCharA->sclr.bColorEnable)
        {
          SendMessage(hWndColorCheck, BM_SETCHECK, BST_CHECKED, 0);
          EnableWindow(hWndColorButton, TRUE);
          EnableWindow(hWndColorLabel, TRUE);
        }
        else
        {
          SendMessage(hWndColorCheck, BM_SETCHECK, BST_UNCHECKED, 0);
          EnableWindow(hWndColorButton, FALSE);
          EnableWindow(hWndColorLabel, FALSE);
        }
        InvalidateRect(hWndColorButton, NULL, FALSE);

        if (pscCharA->sclr.bSelColorEnable)
        {
          SendMessage(hWndSelColorCheck, BM_SETCHECK, BST_CHECKED, 0);
          EnableWindow(hWndSelColorButton, TRUE);
          EnableWindow(hWndSelColorLabel, TRUE);
        }
        else
        {
          SendMessage(hWndSelColorCheck, BM_SETCHECK, BST_UNCHECKED, 0);
          EnableWindow(hWndSelColorButton, FALSE);
          EnableWindow(hWndSelColorLabel, FALSE);
        }
        InvalidateRect(hWndSelColorButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDC_CHAR_EDIT)
    {
      GetWindowTextA(hWndCharEdit, pscCharA->szNewChar, 2);
    }
    else if (LOWORD(wParam) == IDC_COLOR_CHECK)
    {
      if (SendMessage(hWndColorCheck, BM_GETCHECK, 0, 0) == BST_CHECKED)
      {
        EnableWindow(hWndColorButton, TRUE);
        EnableWindow(hWndColorLabel, TRUE);
        pscCharA->sclr.bColorEnable=TRUE;
      }
      else
      {
        EnableWindow(hWndColorButton, FALSE);
        EnableWindow(hWndColorLabel, FALSE);
        pscCharA->sclr.bColorEnable=FALSE;
      }
      InvalidateRect(hWndColorButton, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_SELCOLOR_CHECK)
    {
      if (SendMessage(hWndSelColorCheck, BM_GETCHECK, 0, 0) == BST_CHECKED)
      {
        EnableWindow(hWndSelColorButton, TRUE);
        EnableWindow(hWndSelColorLabel, TRUE);
        pscCharA->sclr.bSelColorEnable=TRUE;
      }
      else
      {
        EnableWindow(hWndSelColorButton, FALSE);
        EnableWindow(hWndSelColorLabel, FALSE);
        pscCharA->sclr.bSelColorEnable=FALSE;
      }
      InvalidateRect(hWndSelColorButton, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_COLOR_BUTTON ||
             LOWORD(wParam) == IDC_SELCOLOR_BUTTON)
    {
      CHOOSECOLORA ccA={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_COLOR_BUTTON)
      {
        lpcrColor=&pscCharA->sclr.dwColor;
        hWndButton=hWndColorButton;
      }
      else if (LOWORD(wParam) == IDC_SELCOLOR_BUTTON)
      {
        lpcrColor=&pscCharA->sclr.dwSelColor;
        hWndButton=hWndSelColorButton;
      }
      ccA.lStructSize  =sizeof(CHOOSECOLORA);
      ccA.hwndOwner    =hDlg;
      ccA.lpCustColors =crCustColors;
      ccA.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccA.rgbResult    =*lpcrColor;

      if (ChooseColorA(&ccA))
      {
        *lpcrColor=ccA.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      scSpaceA=scSpaceDlgA;
      scTabA=scTabDlgA;
      scNewLineA=scNewLineDlgA;
      scVerticalTabA=scVerticalTabDlgA;
      scFormFeedA=scFormFeedDlgA;
      scNullA=scNullDlgA;

      SaveOptionsA();
      EndDialog(hDlg, 0);
      UpdateEditAll();
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  else if (uMsg == WM_DESTROY)
  {
    DestroyIcon(hPluginIcon);
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HICON hPluginIcon;
  static SPECIALCHARW scSpaceDlgW;
  static SPECIALCHARW scTabDlgW;
  static SPECIALCHARW scNewLineDlgW;
  static SPECIALCHARW scVerticalTabDlgW;
  static SPECIALCHARW scFormFeedDlgW;
  static SPECIALCHARW scNullDlgW;
  static SPECIALCHARW *pscCharW;
  static HWND hWndList;
  static HWND hWndCharEdit;
  static HWND hWndColorCheck;
  static HWND hWndColorButton;
  static HWND hWndColorLabel;
  static HWND hWndSelColorCheck;
  static HWND hWndSelColorButton;
  static HWND hWndSelColorLabel;

  if (uMsg == WM_INITDIALOG)
  {
    hPluginIcon=LoadIconW(hInstanceDLL, MAKEINTRESOURCEW(IDI_ICON));
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hPluginIcon);
    hWndList=GetDlgItem(hDlg, IDC_LIST);
    hWndCharEdit=GetDlgItem(hDlg, IDC_CHAR_EDIT);
    hWndColorCheck=GetDlgItem(hDlg, IDC_COLOR_CHECK);
    hWndColorButton=GetDlgItem(hDlg, IDC_COLOR_BUTTON);
    hWndColorLabel=GetDlgItem(hDlg, IDC_COLOR_LABEL);
    hWndSelColorCheck=GetDlgItem(hDlg, IDC_SELCOLOR_CHECK);
    hWndSelColorButton=GetDlgItem(hDlg, IDC_SELCOLOR_BUTTON);
    hWndSelColorLabel=GetDlgItem(hDlg, IDC_SELCOLOR_LABEL);

    SetDlgItemTextW(hDlg, IDC_CHAR_LABEL, GetLangStringW(wLangSystem, STRID_SYMBOL));
    SetDlgItemTextW(hDlg, IDC_COLOR_LABEL, GetLangStringW(wLangSystem, STRID_COLOR));
    SetDlgItemTextW(hDlg, IDC_SELCOLOR_LABEL, GetLangStringW(wLangSystem, STRID_SELCOLOR));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    scSpaceDlgW=scSpaceW;
    scTabDlgW=scTabW;
    scNewLineDlgW=scNewLineW;
    scVerticalTabDlgW=scVerticalTabW;
    scFormFeedDlgW=scFormFeedW;
    scNullDlgW=scNullW;
    pscCharW=&scSpaceDlgW;

    SendMessage(hWndCharEdit, EM_LIMITTEXT, 1, 0);
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_SPACE));
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_TAB));
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_NEWLINE));
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_VERTICALTAB));
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_FORMFEED));
    SendMessageW(hWndList, LB_ADDSTRING, 0, (LPARAM)GetLangStringW(wLangSystem, STRID_NULL));
    SendMessage(hWndList, LB_SETCURSEL, 0, 0);
    SendMessage(hDlg, WM_COMMAND, MAKELONG(IDC_LIST, LBN_SELCHANGE), (LPARAM)hWndList);
  }
  else if (uMsg == WM_DRAWITEM)
  {
    DRAWITEMSTRUCT *dis=(DRAWITEMSTRUCT *)lParam;
    HBRUSH hBrush;
    DWORD dwColor=GetSysColor(COLOR_BTNFACE);

    if (wParam == IDC_COLOR_BUTTON)
    {
      if (pscCharW->sclr.bColorEnable)
        dwColor=pscCharW->sclr.dwColor;
    }
    else if (wParam == IDC_SELCOLOR_BUTTON)
    {
      if (pscCharW->sclr.bSelColorEnable)
        dwColor=pscCharW->sclr.dwSelColor;
    }

    if (hBrush=CreateSolidBrush(dwColor))
    {
      FillRect(dis->hDC, &dis->rcItem, hBrush);
      if (dis->hwndItem == GetFocus())
        DrawFocusRect(dis->hDC, &dis->rcItem);
      DeleteObject(hBrush);
    }
    return TRUE;
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_LIST)
    {
      if (HIWORD(wParam) == LBN_SELCHANGE)
      {
        int nItem;

        nItem=SendMessage(hWndList, LB_GETCURSEL, 0, 0);

        if (nItem == 0)
          pscCharW=&scSpaceDlgW;
        else if (nItem == 1)
          pscCharW=&scTabDlgW;
        else if (nItem == 2)
          pscCharW=&scNewLineDlgW;
        else if (nItem == 3)
          pscCharW=&scVerticalTabDlgW;
        else if (nItem == 4)
          pscCharW=&scFormFeedDlgW;
        else if (nItem == 5)
          pscCharW=&scNullDlgW;

        SetWindowTextW(hWndCharEdit, pscCharW->wszNewChar);

        if (pscCharW->sclr.bColorEnable)
        {
          SendMessage(hWndColorCheck, BM_SETCHECK, BST_CHECKED, 0);
          EnableWindow(hWndColorButton, TRUE);
          EnableWindow(hWndColorLabel, TRUE);
        }
        else
        {
          SendMessage(hWndColorCheck, BM_SETCHECK, BST_UNCHECKED, 0);
          EnableWindow(hWndColorButton, FALSE);
          EnableWindow(hWndColorLabel, FALSE);
        }
        InvalidateRect(hWndColorButton, NULL, FALSE);

        if (pscCharW->sclr.bSelColorEnable)
        {
          SendMessage(hWndSelColorCheck, BM_SETCHECK, BST_CHECKED, 0);
          EnableWindow(hWndSelColorButton, TRUE);
          EnableWindow(hWndSelColorLabel, TRUE);
        }
        else
        {
          SendMessage(hWndSelColorCheck, BM_SETCHECK, BST_UNCHECKED, 0);
          EnableWindow(hWndSelColorButton, FALSE);
          EnableWindow(hWndSelColorLabel, FALSE);
        }
        InvalidateRect(hWndSelColorButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDC_CHAR_EDIT)
    {
      GetWindowTextW(hWndCharEdit, pscCharW->wszNewChar, 2);
    }
    else if (LOWORD(wParam) == IDC_COLOR_CHECK)
    {
      if (SendMessage(hWndColorCheck, BM_GETCHECK, 0, 0) == BST_CHECKED)
      {
        EnableWindow(hWndColorButton, TRUE);
        EnableWindow(hWndColorLabel, TRUE);
        pscCharW->sclr.bColorEnable=TRUE;
      }
      else
      {
        EnableWindow(hWndColorButton, FALSE);
        EnableWindow(hWndColorLabel, FALSE);
        pscCharW->sclr.bColorEnable=FALSE;
      }
      InvalidateRect(hWndColorButton, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_SELCOLOR_CHECK)
    {
      if (SendMessage(hWndSelColorCheck, BM_GETCHECK, 0, 0) == BST_CHECKED)
      {
        EnableWindow(hWndSelColorButton, TRUE);
        EnableWindow(hWndSelColorLabel, TRUE);
        pscCharW->sclr.bSelColorEnable=TRUE;
      }
      else
      {
        EnableWindow(hWndSelColorButton, FALSE);
        EnableWindow(hWndSelColorLabel, FALSE);
        pscCharW->sclr.bSelColorEnable=FALSE;
      }
      InvalidateRect(hWndSelColorButton, NULL, FALSE);
    }
    else if (LOWORD(wParam) == IDC_COLOR_BUTTON ||
             LOWORD(wParam) == IDC_SELCOLOR_BUTTON)
    {
      CHOOSECOLORW ccW={0};
      COLORREF crCustColors[16]={0};
      COLORREF *lpcrColor=NULL;
      HWND hWndButton=NULL;

      if (LOWORD(wParam) == IDC_COLOR_BUTTON)
      {
        lpcrColor=&pscCharW->sclr.dwColor;
        hWndButton=hWndColorButton;
      }
      else if (LOWORD(wParam) == IDC_SELCOLOR_BUTTON)
      {
        lpcrColor=&pscCharW->sclr.dwSelColor;
        hWndButton=hWndSelColorButton;
      }
      ccW.lStructSize  =sizeof(CHOOSECOLORW);
      ccW.hwndOwner    =hDlg;
      ccW.lpCustColors =crCustColors;
      ccW.Flags        =CC_FULLOPEN|CC_RGBINIT;
      ccW.rgbResult    =*lpcrColor;

      if (ChooseColorW(&ccW))
      {
        *lpcrColor=ccW.rgbResult;
        InvalidateRect(hWndButton, NULL, FALSE);
      }
    }
    else if (LOWORD(wParam) == IDOK)
    {
      scSpaceW=scSpaceDlgW;
      scTabW=scTabDlgW;
      scNewLineW=scNewLineDlgW;
      scVerticalTabW=scVerticalTabDlgW;
      scFormFeedW=scFormFeedDlgW;
      scNullW=scNullDlgW;

      SaveOptionsW();
      EndDialog(hDlg, 0);
      UpdateEditAll();
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  else if (uMsg == WM_DESTROY)
  {
    DestroyIcon(hPluginIcon);
  }
  return FALSE;
}

LRESULT CALLBACK NewEditProcA(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_PAINT)
  {
    GETTEXTRANGE gtr;
    CHARRANGE cr;
    TEXTMETRICA tm;
    RECT rcUpdate;
    RECT rcEdit;
    RECT rcDraw;
    HDC hDC;
    HRGN hRgn;
    HRGN hRgnOld;
    HFONT hFont;
    HFONT hFontOld;
    LRESULT lResult;
    int nFirstChar;
    int nLastChar;
    int nFirstCharInUpdateLine;
    int i;

    GetUpdateRect(hWnd, &rcUpdate, FALSE);
    SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
    if (rcUpdate.left < rcEdit.left) rcUpdate.left=rcEdit.left;
    if (rcUpdate.top < rcEdit.top) rcUpdate.top=rcEdit.top;
    if (rcUpdate.right > rcEdit.right) rcUpdate.right=rcEdit.right;
    if (rcUpdate.bottom > rcEdit.bottom) rcUpdate.bottom=rcEdit.bottom;
    rcDraw.left=rcEdit.left;
    rcDraw.top=rcUpdate.top;
    rcDraw.right=rcEdit.right;
    rcDraw.bottom=rcUpdate.bottom;

    lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

    if (hDC=GetDC(hWnd))
    {
      HideCaret(hWnd);
      hRgn=CreateRectRgn(rcEdit.left, rcEdit.top, rcEdit.right, rcEdit.bottom);
      hRgnOld=(HRGN)SelectObject(hDC, hRgn);
      hFont=CreateFontIndirectA((LOGFONTA *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      hFontOld=(HFONT)SelectObject(hDC, hFont);
      GetTextMetricsA(hDC, &tm);
      if (bAkelEdit) tm.tmHeight=SendMessage(hWnd, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
      SetBkMode(hDC, TRANSPARENT);

      SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);
      i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.left);
      i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
      nFirstChar=SendMessage(hWnd, EM_LINEINDEX, i, 0);
      i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.right);
      i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
      i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
      nLastChar=i + SendMessage(hWnd, EM_LINELENGTH, i, 0) + 1;
      gtr.cpMin=nFirstChar;
      gtr.cpMax=nLastChar;
      SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr);

      if (gtr.pText)
      {
        SPECIALCHARA *pscCharA;
        char *szText=(char *)gtr.pText;
        char *pCount;
        POINT ptUpdateLine;
        POINT ptDelimiter;
        int nCharIndex;

        //Get first visible char
        nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcUpdate);
        if (nFirstCharInUpdateLine > nFirstChar) --nFirstCharInUpdateLine;
        if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
        pCount=szText + (nFirstCharInUpdateLine - nFirstChar);

        GetPosFromChar(hWnd, nFirstCharInUpdateLine, &ptUpdateLine, &tm);
        ptUpdateLine.x=rcUpdate.left;

        while (1)
        {
          pscCharA=NULL;
          nCharIndex=nFirstChar + (pCount - szText);
          if (nCharIndex >= nLastChar) break;

          if (*pCount == scSpaceA.szOldChar[0])
            pscCharA=&scSpaceA;
          else if (*pCount == scTabA.szOldChar[0])
            pscCharA=&scTabA;
          else if (*pCount == scNewLineA.szOldChar[0])
            pscCharA=&scNewLineA;
          else if (*pCount == scVerticalTabA.szOldChar[0])
            pscCharA=&scVerticalTabA;
          else if (*pCount == scFormFeedA.szOldChar[0])
            pscCharA=&scFormFeedA;
          else if (*pCount == scNullA.szOldChar[0])
            pscCharA=&scNullA;

          if (pscCharA)
          {
            if (pscCharA->sclr.bColorEnable || pscCharA->sclr.bSelColorEnable)
            {
              GetPosFromChar(hWnd, nCharIndex, &ptDelimiter, &tm);
              if (ptDelimiter.y < ptUpdateLine.y) goto FreeText;

              if (ptDelimiter.y > ptUpdateLine.y ||
                  (ptDelimiter.y == ptUpdateLine.y &&
                   ptDelimiter.x >= rcUpdate.right &&
                   nCharIndex > nFirstCharInUpdateLine))
              {
                ptUpdateLine.y+=tm.tmHeight;

                nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                pCount=szText + (nFirstCharInUpdateLine - nFirstChar);
                continue;
              }
              else
              {
                if (ptDelimiter.x + tm.tmMaxCharWidth > rcUpdate.left)
                {
                  rcDraw.left=ptDelimiter.x;
                  rcDraw.top=ptDelimiter.y;

                  if (!(nCharIndex >= cr.cpMin && nCharIndex < cr.cpMax))
                  {
                    if (pscCharA->sclr.bColorEnable)
                    {
                      CHARCOLOR cc;

                      //Erase old
                      if (pscCharA->szOldChar[0] != '\r' &&
                          pscCharA->szOldChar[0] != '\t')
                      {
                        cc.nCharPos=nCharIndex;
                        SendMessage(hMainWnd, AKD_GETCHARCOLOR, (WPARAM)hWnd, (LPARAM)&cc);
                        SetTextColor(hDC, cc.crBk);
                        TextOutA(hDC, rcDraw.left, rcDraw.top, pscCharA->szOldChar, 1);
                      }

                      //Draw new
                      SetTextColor(hDC, pscCharA->sclr.dwColor);
                      TextOutA(hDC, rcDraw.left, rcDraw.top, pscCharA->szNewChar, 1);
                    }
                  }
                  else
                  {
                    if (pscCharA->sclr.bSelColorEnable)
                    {
                      CHARCOLOR cc;

                      //Erase old
                      if (pscCharA->szOldChar[0] != '\r' &&
                          pscCharA->szOldChar[0] != '\t')
                      {
                        cc.nCharPos=nCharIndex;
                        SendMessage(hMainWnd, AKD_GETCHARCOLOR, (WPARAM)hWnd, (LPARAM)&cc);
                        SetTextColor(hDC, cc.crBk);
                        TextOutA(hDC, rcDraw.left, rcDraw.top, pscCharA->szOldChar, 1);
                      }

                      //Draw new
                      SetTextColor(hDC, pscCharA->sclr.dwSelColor);
                      TextOutA(hDC, rcDraw.left, rcDraw.top, pscCharA->szNewChar, 1);
                    }
                  }
                }
              }
            }
          }
          ++pCount;
        }
        FreeText:
        SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
      }
      if (hFontOld) SelectObject(hDC, hFontOld);
      DeleteObject(hFont);
      if (hRgnOld) SelectObject(hDC, hRgnOld);
      DeleteObject(hRgn);
      ShowCaret(hWnd);
      ReleaseDC(hWnd, hDC);
    }
    return lResult;
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewEditProcW(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_PAINT)
  {
    GETTEXTRANGE gtr;
    CHARRANGE cr;
    TEXTMETRICW tm;
    RECT rcUpdate;
    RECT rcEdit;
    RECT rcDraw;
    HDC hDC;
    HRGN hRgn;
    HRGN hRgnOld;
    HFONT hFont;
    HFONT hFontOld;
    LRESULT lResult;
    int nFirstChar;
    int nLastChar;
    int nFirstCharInUpdateLine;
    int i;

    GetUpdateRect(hWnd, &rcUpdate, FALSE);
    SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcEdit);
    if (rcUpdate.left < rcEdit.left) rcUpdate.left=rcEdit.left;
    if (rcUpdate.top < rcEdit.top) rcUpdate.top=rcEdit.top;
    if (rcUpdate.right > rcEdit.right) rcUpdate.right=rcEdit.right;
    if (rcUpdate.bottom > rcEdit.bottom) rcUpdate.bottom=rcEdit.bottom;
    rcDraw.left=rcEdit.left;
    rcDraw.top=rcUpdate.top;
    rcDraw.right=rcEdit.right;
    rcDraw.bottom=rcUpdate.bottom;

    lResult=NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);

    if (hDC=GetDC(hWnd))
    {
      HideCaret(hWnd);
      hRgn=CreateRectRgn(rcEdit.left, rcEdit.top, rcEdit.right, rcEdit.bottom);
      hRgnOld=(HRGN)SelectObject(hDC, hRgn);
      hFont=CreateFontIndirectW((LOGFONTW *)SendMessage(hMainWnd, AKD_GETFONT, (WPARAM)hWnd, 0));
      hFontOld=(HFONT)SelectObject(hDC, hFont);
      GetTextMetricsW(hDC, &tm);
      if (bAkelEdit) tm.tmHeight=SendMessage(hWnd, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
      SetBkMode(hDC, TRANSPARENT);

      SendMessage(hWnd, EM_EXGETSEL, 0, (LPARAM)&cr);
      i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.left);
      i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
      nFirstChar=SendMessage(hWnd, EM_LINEINDEX, i, 0);
      i=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcDraw.right);
      i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, i);
      i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
      nLastChar=i + SendMessage(hWnd, EM_LINELENGTH, i, 0) + 1;
      gtr.cpMin=nFirstChar;
      gtr.cpMax=nLastChar;
      SendMessage(hMainWnd, AKD_GETTEXTRANGE, (WPARAM)hWnd, (LPARAM)&gtr);

      if (gtr.pText)
      {
        SPECIALCHARW *pscCharW;
        wchar_t *wszText=(wchar_t *)gtr.pText;
        wchar_t *wpCount;
        POINT ptUpdateLine;
        POINT ptDelimiter;
        int nCharIndex;

        //Get first visible char
        nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&rcUpdate);
        if (nFirstCharInUpdateLine > nFirstChar) --nFirstCharInUpdateLine;
        if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
        wpCount=wszText + (nFirstCharInUpdateLine - nFirstChar);

        GetPosFromChar(hWnd, nFirstCharInUpdateLine, &ptUpdateLine, (TEXTMETRICA *)&tm);
        ptUpdateLine.x=rcUpdate.left;

        while (1)
        {
          pscCharW=NULL;
          nCharIndex=nFirstChar + (wpCount - wszText);
          if (nCharIndex >= nLastChar) break;

          if (*wpCount == scSpaceW.wszOldChar[0])
            pscCharW=&scSpaceW;
          else if (*wpCount == scTabW.wszOldChar[0])
            pscCharW=&scTabW;
          else if (*wpCount == scNewLineW.wszOldChar[0])
            pscCharW=&scNewLineW;
          else if (*wpCount == scVerticalTabW.wszOldChar[0])
            pscCharW=&scVerticalTabW;
          else if (*wpCount == scFormFeedW.wszOldChar[0])
            pscCharW=&scFormFeedW;
          else if (*wpCount == scNullW.wszOldChar[0])
            pscCharW=&scNullW;

          if (pscCharW)
          {
            if (pscCharW->sclr.bColorEnable || pscCharW->sclr.bSelColorEnable)
            {
              GetPosFromChar(hWnd, nCharIndex, &ptDelimiter, (TEXTMETRICA *)&tm);
              if (ptDelimiter.y < ptUpdateLine.y) goto FreeText;

              if (ptDelimiter.y > ptUpdateLine.y ||
                  (ptDelimiter.y == ptUpdateLine.y &&
                   ptDelimiter.x >= rcUpdate.right &&
                   nCharIndex > nFirstCharInUpdateLine))
              {
                ptUpdateLine.y+=tm.tmHeight;

                nFirstCharInUpdateLine=SendMessage(hWnd, EM_CHARFROMPOS, 0, (LPARAM)&ptUpdateLine);
                i=SendMessage(hWnd, EM_EXLINEFROMCHAR, 0, nFirstCharInUpdateLine);
                i=SendMessage(hWnd, EM_LINEINDEX, i, 0);
                if (nFirstCharInUpdateLine > i) --nFirstCharInUpdateLine;
                if (nFirstCharInUpdateLine >= nLastChar) goto FreeText;
                wpCount=wszText + (nFirstCharInUpdateLine - nFirstChar);
                continue;
              }
              else
              {
                if (ptDelimiter.x + tm.tmMaxCharWidth > rcUpdate.left)
                {
                  rcDraw.left=ptDelimiter.x;
                  rcDraw.top=ptDelimiter.y;

                  if (!(nCharIndex >= cr.cpMin && nCharIndex < cr.cpMax))
                  {
                    if (pscCharW->sclr.bColorEnable)
                    {
                      CHARCOLOR cc;

                      //Erase old
                      if (pscCharW->wszOldChar[0] != L'\r' &&
                          pscCharW->wszOldChar[0] != L'\t')
                      {
                        cc.nCharPos=nCharIndex;
                        SendMessage(hMainWnd, AKD_GETCHARCOLOR, (WPARAM)hWnd, (LPARAM)&cc);
                        SetTextColor(hDC, cc.crBk);
                        TextOutW(hDC, rcDraw.left, rcDraw.top, pscCharW->wszOldChar, 1);
                      }

                      //Draw new
                      SetTextColor(hDC, pscCharW->sclr.dwColor);
                      TextOutW(hDC, rcDraw.left, rcDraw.top, pscCharW->wszNewChar, 1);
                    }
                  }
                  else
                  {
                    if (pscCharW->sclr.bSelColorEnable)
                    {
                      CHARCOLOR cc;

                      //Erase old
                      if (pscCharW->wszOldChar[0] != L'\r' &&
                          pscCharW->wszOldChar[0] != L'\t')
                      {
                        cc.nCharPos=nCharIndex;
                        SendMessage(hMainWnd, AKD_GETCHARCOLOR, (WPARAM)hWnd, (LPARAM)&cc);
                        SetTextColor(hDC, cc.crBk);
                        TextOutW(hDC, rcDraw.left, rcDraw.top, pscCharW->wszOldChar, 1);
                      }

                      //Draw new
                      SetTextColor(hDC, pscCharW->sclr.dwSelColor);
                      TextOutW(hDC, rcDraw.left, rcDraw.top, pscCharW->wszNewChar, 1);
                    }
                  }
                }
              }
            }
          }
          ++wpCount;
        }
        FreeText:
        SendMessage(hMainWnd, AKD_FREETEXT, 0, (LPARAM)gtr.pText);
      }
      if (hFontOld) SelectObject(hDC, hFontOld);
      DeleteObject(hFont);
      if (hRgnOld) SelectObject(hDC, hRgnOld);
      DeleteObject(hRgn);
      ShowCaret(hWnd);
      ReleaseDC(hWnd, hDC);
    }
    return lResult;
  }

  //Call next procedure
  return NewEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void UpdateEdit(HWND hWnd)
{
  RECT rcDraw;

  SendMessage(hWnd, EM_GETRECT, 0, (LPARAM)&rcDraw);
  InvalidateRect(hWnd, &rcDraw, FALSE);
}

void UpdateEditAll()
{
  HWND hWndFrame;
  HWND hWndChild;

  if (!bMDI)
  {
    hWndChild=GetWindow(hMainWnd, GW_CHILD);

    while (hWndChild)
    {
      if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
        UpdateEdit(hWndChild);
      hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
    }
  }
  else
  {
    hWndFrame=GetWindow(hMdiClient, GW_CHILD);

    while (hWndFrame)
    {
      hWndChild=GetWindow(hWndFrame, GW_CHILD);

      while (hWndChild)
      {
        if (GetWindowLongA(hWndChild, GWL_ID) == ID_EDIT)
          UpdateEdit(hWndChild);
        hWndChild=GetNextWindow(hWndChild, GW_HWNDNEXT);
      }
      hWndFrame=GetNextWindow(hWndFrame, GW_HWNDNEXT);
    }
  }
}

void GetPosFromChar(HWND hWnd, int nCharIndex, POINT *pt, TEXTMETRICA *tm)
{
  DWORD dwPosition;

  if (bOldRichEdit)
  {
    dwPosition=SendMessage(hWnd, EM_POSFROMCHAR, nCharIndex, (LPARAM)NULL);
    pt->x=LOWORD(dwPosition);
    pt->y=HIWORD(dwPosition);

    if (tm)
    {
      if (pt->x > 0x10000 - tm->tmMaxCharWidth)
        pt->x-=0x10000;
      if (pt->y > 0x10000 - tm->tmHeight)
        pt->y-=0x10000;
    }
  }
  else
  {
    SendMessage(hWnd, EM_POSFROMCHAR, (WPARAM)pt, nCharIndex);
  }
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="SpaceChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scSpaceA.szNewChar;
    po.dwData=sizeof(scSpaceA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SpaceColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scSpaceA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="TabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scTabA.szNewChar;
    po.dwData=sizeof(scTabA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="TabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scTabA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NewLineChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNewLineA.szNewChar;
    po.dwData=sizeof(scNewLineA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NewLineColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNewLineA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="VerticalTabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scVerticalTabA.szNewChar;
    po.dwData=sizeof(scVerticalTabA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="VerticalTabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scVerticalTabA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="FormFeedChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scFormFeedA.szNewChar;
    po.dwData=sizeof(scFormFeedA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="FormFeedColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scFormFeedA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NullChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNullA.szNewChar;
    po.dwData=sizeof(scNullA.szNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NullColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNullA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"SpaceChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scSpaceW.wszNewChar;
    po.dwData=sizeof(scSpaceW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SpaceColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scSpaceW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"TabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scTabW.wszNewChar;
    po.dwData=sizeof(scTabW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"TabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scTabW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NewLineChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNewLineW.wszNewChar;
    po.dwData=sizeof(scNewLineW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NewLineColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNewLineW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"VerticalTabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scVerticalTabW.wszNewChar;
    po.dwData=sizeof(scVerticalTabW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"VerticalTabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scVerticalTabW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"FormFeedChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scFormFeedW.wszNewChar;
    po.dwData=sizeof(scFormFeedW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"FormFeedColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scFormFeedW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NullChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNullW.wszNewChar;
    po.dwData=sizeof(scNullW.wszNewChar);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NullColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNullW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    scSpaceA.szNewChar[1]='\0';
    po.pOptionName="SpaceChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scSpaceA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="SpaceColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scSpaceA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scTabA.szNewChar[1]='\0';
    po.pOptionName="TabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scTabA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="TabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scTabA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scNewLineA.szNewChar[1]='\0';
    po.pOptionName="NewLineChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNewLineA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NewLineColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNewLineA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scVerticalTabA.szNewChar[1]='\0';
    po.pOptionName="VerticalTabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scVerticalTabA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="VerticalTabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scVerticalTabA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scFormFeedA.szNewChar[1]='\0';
    po.pOptionName="FormFeedChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scFormFeedA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="FormFeedColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scFormFeedA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scNullA.szNewChar[1]='\0';
    po.pOptionName="NullChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNullA.szNewChar;
    po.dwData=2;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NullColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNullA.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    scSpaceW.wszNewChar[1]='\0';
    po.pOptionName=L"SpaceChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scSpaceW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"SpaceColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scSpaceW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scTabW.wszNewChar[1]='\0';
    po.pOptionName=L"TabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scTabW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"TabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scTabW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scNewLineW.wszNewChar[1]='\0';
    po.pOptionName=L"NewLineChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNewLineW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NewLineColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNewLineW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scVerticalTabW.wszNewChar[1]='\0';
    po.pOptionName=L"VerticalTabChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scVerticalTabW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"VerticalTabColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scVerticalTabW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scFormFeedW.wszNewChar[1]='\0';
    po.pOptionName=L"FormFeedChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scFormFeedW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"FormFeedColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scFormFeedW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    scNullW.wszNewChar[1]='\0';
    po.pOptionName=L"NullChar";
    po.dwType=PO_STRING;
    po.lpData=(LPBYTE)scNullW.wszNewChar;
    po.dwData=4;
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NullColor";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&scNullW.sclr;
    po.dwData=sizeof(SPECIALCOLOR);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
    if (nStringID == STRID_SYMBOL)
      return L"\x0421\x0438\x043C\x0432\x043E\x043B";
    if (nStringID == STRID_COLOR)
      return L"\x0426\x0432\x0435\x0442";
    if (nStringID == STRID_SELCOLOR)
      return L"\x0426\x0432\x0435\x0442\x0020\x0432\x0020\x0432\x044B\x0434\x0435\x043B\x0435\x043D\x0438\x0438";
    if (nStringID == STRID_SPACE)
      return L"\x041F\x0440\x043E\x0431\x0435\x043B";
    if (nStringID == STRID_TAB)
      return L"\x0422\x0430\x0431\x0443\x043B\x044F\x0446\x0438\x044F";
    if (nStringID == STRID_NEWLINE)
      return L"\x041D\x043E\x0432\x0430\x044F\x0020\x0441\x0442\x0440\x043E\x043A\x0430";
    if (nStringID == STRID_VERTICALTAB)
      return L"\x0412\x0435\x0440\x0442\x0438\x043A\x0430\x043B\x044C\x043D\x0430\x044F\x0020\x0442\x0430\x0431\x0443\x043B\x044F\x0446\x0438\x044F";
    if (nStringID == STRID_FORMFEED)
      return L"\x041F\x0440\x043E\x0433\x043E\x043D\x0020\x043B\x0438\x0441\x0442\x0430";
    if (nStringID == STRID_NULL)
      return L"\x041D\x0443\x043B\x0435\x0432\x043E\x0439\x0020\x0441\x0438\x043C\x0432\x043E\x043B";
  }
  else
  {
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
    if (nStringID == STRID_SYMBOL)
      return L"Symbol";
    if (nStringID == STRID_COLOR)
      return L"Color";
    if (nStringID == STRID_SELCOLOR)
      return L"Color in selection";
    if (nStringID == STRID_SPACE)
      return L"Space";
    if (nStringID == STRID_TAB)
      return L"Tabulation";
    if (nStringID == STRID_NEWLINE)
      return L"New line";
    if (nStringID == STRID_VERTICALTAB)
      return L"Vertical tabulation";
    if (nStringID == STRID_FORMFEED)
      return L"Form-feed";
    if (nStringID == STRID_NULL)
      return L"Null";
  }
  return L"";
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  bOldWindows=pd->bOldWindows;
  bOldRichEdit=pd->bOldRichEdit;
  bAkelEdit=pd->bAkelEdit;
  bMDI=pd->bMDI;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    scSpaceA.szNewChar[0]='\xB7';
    scSpaceA.szNewChar[1]='\0';
    scSpaceA.szOldChar[0]=' ';
    scSpaceA.szOldChar[1]='\0';
    scSpaceA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scSpaceA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scSpaceA.sclr.bColorEnable=TRUE;
    scSpaceA.sclr.bSelColorEnable=TRUE;

    scTabA.szNewChar[0]='\xBB';
    scTabA.szNewChar[1]='\0';
    scTabA.szOldChar[0]='\t';
    scTabA.szOldChar[1]='\0';
    scTabA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scTabA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scTabA.sclr.bColorEnable=TRUE;
    scTabA.sclr.bSelColorEnable=TRUE;

    scNewLineA.szNewChar[0]='\xB6';
    scNewLineA.szNewChar[1]='\0';
    scNewLineA.szOldChar[0]='\r';
    scNewLineA.szOldChar[1]='\0';
    scNewLineA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scNewLineA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scNewLineA.sclr.bColorEnable=TRUE;
    scNewLineA.sclr.bSelColorEnable=TRUE;

    scVerticalTabA.szNewChar[0]='\xA6';
    scVerticalTabA.szNewChar[1]='\0';
    scVerticalTabA.szOldChar[0]='\v';
    scVerticalTabA.szOldChar[1]='\0';
    scVerticalTabA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scVerticalTabA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scVerticalTabA.sclr.bColorEnable=TRUE;
    scVerticalTabA.sclr.bSelColorEnable=TRUE;

    scFormFeedA.szNewChar[0]='\xA7';
    scFormFeedA.szNewChar[1]='\0';
    scFormFeedA.szOldChar[0]='\f';
    scFormFeedA.szOldChar[1]='\0';
    scFormFeedA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scFormFeedA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scFormFeedA.sclr.bColorEnable=TRUE;
    scFormFeedA.sclr.bSelColorEnable=TRUE;

    scNullA.szNewChar[0]='\xA4';
    scNullA.szNewChar[1]='\0';
    scNullA.szOldChar[0]='\0';
    scNullA.szOldChar[1]='\0';
    scNullA.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scNullA.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scNullA.sclr.bColorEnable=TRUE;
    scNullA.sclr.bSelColorEnable=TRUE;

    ReadOptionsA();
  }
  else
  {
    scSpaceW.wszNewChar[0]=L'\x00B7';
    scSpaceW.wszNewChar[1]=L'\0';
    scSpaceW.wszOldChar[0]=L' ';
    scSpaceW.wszOldChar[1]=L'\0';
    scSpaceW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scSpaceW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scSpaceW.sclr.bColorEnable=TRUE;
    scSpaceW.sclr.bSelColorEnable=TRUE;

    scTabW.wszNewChar[0]=L'\x00BB';
    scTabW.wszNewChar[1]=L'\0';
    scTabW.wszOldChar[0]=L'\t';
    scTabW.wszOldChar[1]=L'\0';
    scTabW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scTabW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scTabW.sclr.bColorEnable=TRUE;
    scTabW.sclr.bSelColorEnable=TRUE;

    scNewLineW.wszNewChar[0]=L'\x00B6';
    scNewLineW.wszNewChar[1]=L'\0';
    scNewLineW.wszOldChar[0]=L'\r';
    scNewLineW.wszOldChar[1]=L'\0';
    scNewLineW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scNewLineW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scNewLineW.sclr.bColorEnable=TRUE;
    scNewLineW.sclr.bSelColorEnable=TRUE;

    scVerticalTabW.wszNewChar[0]=L'\x00A6';
    scVerticalTabW.wszNewChar[1]=L'\0';
    scVerticalTabW.wszOldChar[0]=L'\v';
    scVerticalTabW.wszOldChar[1]=L'\0';
    scVerticalTabW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scVerticalTabW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scVerticalTabW.sclr.bColorEnable=TRUE;
    scVerticalTabW.sclr.bSelColorEnable=TRUE;

    scFormFeedW.wszNewChar[0]=L'\x00A7';
    scFormFeedW.wszNewChar[1]=L'\0';
    scFormFeedW.wszOldChar[0]=L'\f';
    scFormFeedW.wszOldChar[1]=L'\0';
    scFormFeedW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scFormFeedW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scFormFeedW.sclr.bColorEnable=TRUE;
    scFormFeedW.sclr.bSelColorEnable=TRUE;

    scNullW.wszNewChar[0]=L'\x00A4';
    scNullW.wszNewChar[1]=L'\0';
    scNullW.wszOldChar[0]=L'\0';
    scNullW.wszOldChar[1]=L'\0';
    scNullW.sclr.dwColor=RGB(0x33, 0x33, 0xCC);
    scNullW.sclr.dwSelColor=GetSysColor(COLOR_HIGHLIGHTTEXT);
    scNullW.sclr.bColorEnable=TRUE;
    scNullW.sclr.bSelColorEnable=TRUE;

    ReadOptionsW();
  }
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  if (bOldWindows)
  {
    NewEditProcData=NULL;
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProcA, (LPARAM)&NewEditProcData);
  }
  else
  {
    NewEditProcData=NULL;
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NewEditProcW, (LPARAM)&NewEditProcData);
  }
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove subclass
  if (NewEditProcData)
  {
    SendMessage(hMainWnd, AKD_SETEDITPROC, (WPARAM)NULL, (LPARAM)&NewEditProcData);
    NewEditProcData=NULL;
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
